
; ---------------------------------------------------------------------------------------------
; BIBLIOTECA DE FUNCOES ARITMETICAS DE 32 BITS PARA FAMILIA 18F           Original - 29/07/2011
;
; STATUS: FUNCIONA
;
; ---------------------------------------------------------------------------------------------
; Este modulo fornece funcoes aritmeticas de 32 bits para microcontroladores da familia 18F
; permitindo que os registradores aritmeticos residam em qualquer posicao da RAM gracas ao uso
; de enderecamento indireto.
;
; Este servico permite que os registradores indexadores FSRx sejam salvos no stack de software.
; Para que os registradores indexadores FSRx sejam salvos no stack, todas as funcoes devem ser
; chamadas atraves das macroinstrucoes ADD, SUBTRACT, DIVIDE, MULTIPLY, SQRT, BIN2DEC, DEC2BIN,
; NEGATE, CLRA, CLRB, CLRDIG, MATHLD, MATHLR, MATHST e MATHASCII. Porem, antes de usar essas
; macros, deve-se definir o simbolo '__STACK_FSR' no programa para executar as funcoes em modo
; de salvamento.
;
; Se as funcoes forem chamadas diretamente pela instrucao CALL, os registradores indexadores 
; FSRx nao serao salvos no stack.
; 
; Este servico e' composto dos seguintes arquivos:
;  Registradores aritmeticos ......................... #include "..\..\Includes\math.inc"
;  Funcao de comparacao .............................. #include "..\..\Services\mathcrf_services.inc"
;  Funcoes aritmeticas e macros ...................... #include "..\..\Services\math_services.inc"
;  Facilidades ....................................... #include "..\..\Services\mathfacility_services.inc"
;
; ---------------------------------------------------------------------------------------------
; NOTAS
; 1 - O arquivo "mathcrf_services.inc" contem uma funcao exclusivamente para uso interno que nao
;     deve ser usada pelo programador. Esta funcao e' usada pelas macros CRF e CRFI para comparar
;     registradores de 32 bits.
;
; ---------------------------------------------------------------------------------------------
; Este servico requer que os seguintes arquivos sejam incluidos no codigo fonte:
;
;   #include "..\..\Includes\math.inc"
;         Definicao de registradores na RAM.
;         os registradores podem residir em qualquer
;         posicao da RAM. Por exemplo:
;         MATHRAM   UDATA   0x400            ; <-- registradores alocados na posicao 0x400 da RAM
;          #include "..\..\Includes\math.inc" 
;
;  #include "..\..\Services\math_services,inc"
;        Este arquivo contem o codigo e as macros instrucoes das funcoes aritmetica. 
;
;  #include "..\..\Services\stack_services.inc"
;        Este arquivo contem os servicos de stack de software.
;
; ---------------------------------------------------------------------------------------------
;   MODIFICACOES:
;
;   21/12/2017 -  Os registradores SVFSR0L, SVFSR0H, SVFSR1L, SVFSR1H, SVFSR2L e SVFSR2H foram 
;                 eliminados para poupar memoria RAM, pois nenhuma funcao fazia referencia a 
;                 eles. Na funcao 'dec2bin', o registrador SVStatus foi substituido por REGC3.
;
;   12/11/2017 - As macros CR e CRI foram eliminadas.
;                Foram criadas as macros CRF e CRFI para comparacao de registradores de 32 bits
;                em qualquer localizacao da RAM, usando enderecamento indireto (FSR1/FSR2) e 
;                com opcao de salvamento dos registradores FSR0, FSR1 e FSR2 no stack de software. 
;                A funcao interna '_saveFSR' foi descontinuada em virtude da criacao da macro
;                STACK_FSR. A macro STACK_FSR solicita a opcao de salvamento na execucao das 
;                seguintes macros: ADD, SUBTRACT, DIVIDE, MULTIPLY, SQRT, BIN2DEC, DEC2BIN, 
;                NEGATE, CLRA, CLRB, CLRDIG, MATHLD, MATHLR, MATHST e MATHASCII.
;                As macros BE, BL e BH ainda sao as mesmas para fazer desvios apos a comparacao.  
;
;   11/11/2017 - Foram criadas as macros CR e CRI para permitir comparacoes entre registradores 
;                de 32 bits somente no banco zero da RAM e tambem as macros BE, BL e BH para 
;                fazer desvios apos a comparacao.  
;
;   31/12/2015 - A partir desta versao os registradores FSR0, FSR1, e FRS2 sao salvos
;                ao entrar nas funcoes e restaurados ao sair. Para isso acontecer e'
;                necessario chamar as funcoes atraves das macros MATHLD, MATHLR, MATHST, 
;                MATHASCII, ADD, SUBTRACT, DIVIDE, MULTIPLY, SQRT, BIN2DEC, DEC2BIN, NEGATE, 
;                CLRA, CLRB e CLRDIG.
;
;   30/12/2015 - Alteracao para usar enderecamento indireto permitindo trabalhar em qualquer
;                localizacao da RAM e criacao de macros auxiliares para carga de registradores
;                comuns e registradores aritmeticos.
;                A partir desta versao este modulo passa a usar os registradores FSR0, FSR1 e
;                FSR2 mas nao faz o salvamento.
;
; ---------------------------------------------------------------------------------------------


  if (__FAMILIA==16 ||  __FAMILIA==12) 
     messg "Carregando math16_services.inc..."
     #include "..\..\Services\math16_services.inc"
#define  __MATH_SERVICES 1  ; para nao gerar o codigo abaixo
  endif




  ifndef __MATH_SERVICES
#define  __MATH_SERVICES 1


; -----------------------------------------------------------------------------------------------------------
; FUNCOES DEFINIDAS NESTE SERVICO:
;
; subtract........................ REGA - REGB -> REGA
; add............................. REGA + REGB -> REGA
; multiply........................ REGA * REGB -> REGA
; divide.......................... REGA / REGB -> REGA   Resto em REGC
; sqrt............................ sqrt(REGA) -> REGA
; bin2dec......................... REGA -> DIGITS 1 (MSD) TO 10 (LSD) & DSIGN
; dec2bin......................... Decimal DIGIT1 a DIGIT(X) & DSIGN -> REGA  W deve conter tamanho
; clra............................ limpa REGA0 a REGA3
; clrb............................ limpa REGB0 a REGB3
; clrc............................ limpa REGC0 a REGC3
; clrdig.......................... limpa DIGIT1 a DIGIT10
;
; -----------------------------------------------------------------------------------------------------------

; -----------------------------------------------------------------------------------------------------------
; MACROS DE FUNCOES ARITMETICAS                                                                    12/11/2017
; 
; As macros abaixo executam funcoes de controle para operacoes aritmeticas.
;
; As macros marcadas com [mathfacility] pertencem ao arquivo mathfacility_services.inc.
;
; -----------------------------------------------------------------------------------------------------------
; ADD                REGA = REGA + REGB
;
; AR  r1, r2         [mathfacility]
;                    Soma de registradores de 32 bits.
;                    r1 = r1 + r2 
;                    O registrador r1 contera 0 se positivo ou 1 se negativo.
;                
; SR  r1, r2         [mathfacility]
;                    Subtracao de registradores de 32 bits.
;                    r1 = r1 - r2 
;                    O registrador r1 contera 0 se positivo ou 1 se negativo.
;                
; BE  destino        Apos comparacao desvia se Op1 == Op2.
;
; BH  destino        Apos comparacao desvia se Op1 > Op2.
;
; BIN2DEC            Converte binario em REGA para decimal em DIGIT1 a DIGIT10.
;
; BL  destino        Apos comparacao desvia se Op1 < Op2.
;
; BON r1, destino    [mathfacility]
;                    Desvia se o bit 7 de r1 de 8 bits estiver ligado (numero negativo).
;
; BOP r1, destino    [mathfacility]
;                    Desvia se o bit 7 de r1 de 8 bits estiver desligado (numero positivo).
;
; BOZ r1, destino    [mathfacility]
;                    Desvia se o registrador r1 de 32 bits for zero.
;
;
;
; CLRA               Zera REGA.
;
; CLRB               Zera REGB.
;
; CLRDIG             Zera todos os registradores DIGIT.
;
; CRF rx, ry         Compara 2 registradores de 32 bits (rx e ry) em qualquer lugar da RAM, salvando
;                    os conteudos de FSR0, FSR1 e FSR2 se assim for desejado.
;                    O todos os dois operandos podem residir em qualquer lugar da RAM.
;                    A macro STACK_FSR oferece opcao de salvar no stack de software, os 
;                    registradores de trabalho REGCx, FSR1 e FSR2. 
;
;                    NOTAS:
;                    1 - Os registradores indexadores FSR1 e FSR2 poderao ser destruidos ou
;                        preservados com o uso da macro STACK_ALLFSR_MATH_REGS. 
;                    2 - Os registradores aritmeticos REGC3 a REGC0 poderao ser destruidos ou
;                        preservados com o uso da macro STACK_ALLFSR_MATH_REGS. 
;                    3 - Os registradores aritmeticos MCOUNT e DCOUNT serao destruidos.
;                    4 - O resultado da comparacao estara nos flags de STATUS Z e C.
;                        Z      C   Resultado
;                        0      0   Op1 < Op2   Negativo
;                        0      1   Op1 > Op2   Positivo
;                        1      1   Op1 = Op2   Nulo
;                    5 - Para fazer desvios apos uma comparacao, use as macros BE, BL e BH definidas neste
;                        arquivo.
;                    6 - Os registradores REGC3, REGC2, REGC1 e REGC0 serao destruidos.
; 
;
; CRFI rx, valor     Compara o registrador de 32 bits no primeiro operando com valor imediato, salvando
;                    os conteudos de FSR0, FSR1 e FSR2 se assim for desejado.
;                    O primeiro operando pode residir em qualquer lugar da RAM.
;                    O segundo operando deve ser um valor imediato, isto e', como 0x012286.
;                    A macro STACK_FSR oferece opcao de salvar no stack de software, os registradores de
;                    trabalho REGCx, FSR1 e FSR2. 
;                    Apesar do registrador que esta sendo comparado ser de 32 bits, apenas seus 24 bits
;                    de mais baixa ordem serao comparados com o valor imediato devido a limitacao da
;                    linguagem assembler nao ter como carregar o valor do byte de mais alta ordem, isto
;                    e', so podemos carregar valores de 24 bits com UPPER(x), HIGH(x) e LOW(x).
;
;                    NOTAS:
;                    1 - Os registradores indexadores FSR1 e FSR2 poderao ser destruidos ou
;                        preservados com o uso da macro STACK_ALLFSR_MATH_REGS. 
;                    2 - Os registradores aritmeticos REGC3 a REGC0 poderao ser destruidos ou
;                        preservados com o uso da macro STACK_ALLFSR_MATH_REGS. 
;                    3 - Os registradores aritmeticos MCOUNT e DCOUNT serao destruidos.
;                    4 - O resultado da comparacao estara nos flags de STATUS Z e C.
;                        Z      C   Resultado
;                        0      0   Op1 < Op2   Negativo
;                        0      1   Op1 > Op2   Positivo
;                        1      1   Op1 = Op2   Nulo
;                    5 - Para fazer desvios apos uma comparacao, use as macros BE, BL e BH definidas neste
;                        arquivo.
;                    6 - Os registradores REGC3, REGC2, REGC1 e REGC0 serao destruidos.
;
;
; DEC2BIN            Converte decimal em DIGIT1 a DIGITn para binario em REGA.
;
; DIVIDE             REGA = REGA / REGB  resto em REGC
;
; INCR r1            [mathfacility]
;                    Incrementa (soma 1) um registrador de 32 bits.
;
; LFR f1, r1         [mathfacility]
;                    Carrega o registrador f1 de 8 bits com o valor do byte menos significativo (LSB)
;                    do registrador de 32 bits r1.
;
;
; LR  r1, r2         [mathfacility]
;                    Carrega o valor de 32 bits de r2 no registrador r1 
;
; LWF r1             [mathfacility]
;                    Carrega o registrador  W com o valor do byte menos significativo (LSB) do 
;                    registrador de 32 bits r1.
;
; MATHASCII          Insere 0x30 no nible MSB nos registradores DIGIT1 a DIGIT10 apos uma conversao de
;                    binario para decimal (bin2dec) a fim de  transformar os valores desses registradores 
;                    em formato ASCII.
;
; MATHLD  REGx, val  Carrega um registrador aritmetico (32bits) com um valor imediato de 24 bits.
;                    O primeiro operando e' o registrador aritmetico.
;                    O segundo operando e' o valor imediato que devera estar entre 0 (zero)
;                    e 16.777.215 (0xFFFFFF) no maximo, inclusive. 
;                    O valor imediato pode ser negativo.
;
; MATHLR REGx, rg    Armazena um registrador qualquer de 32 bits do segundo operando num
;                    registrador aritmetico do primeiro operando.
;
; MATHST REGx, rg    Armazena o registrador aritmetico de 32 bits do primeiro operando no 
;                    registrador qualquer de 32 bits do segundo operando.
;
; MR  r1, r2         [mathfacility]
;                    Multiplicacao de registradores de 32 bits.
;                    r1 = r1 * r2 
;                
; MULTIPLY           REGA = REGA * REGB
;
; NEGATE             REGx = -REGx    sendo x=A ou x=B.
;
; PACK               PACK  orig, digits, dest            (27/12/2019)
;                    Compacta o numero decimal no formato 0N0N...0N, contido no registrador 'orig',
;                    para o formato BCD no formato NN...NN no registrado 'dest' com 'digits' digitos
;                    decimais. Se o registrador 'orig' contiveer um numero impar de digitos, o MSB
;                    do resultado no registrador 'dest' sera preenchido com zero.
;                    Para descompactar, usar a macro UNPACK.
;
; RL                 [mathfacility]
;                    Rotaciona para a esquerda os bits de um registrador de tamanho variavel.
;                    O registrador pode residir em qualquer banco da RAM.
;
; RR                 [mathfacility]
;                    Rotaciona para a direita os bits de um registrador de tamanho variavel.
;                    O registrador pode residir em qualquer banco da RAM.
;
; SQRT               REGA = sqrt(REGA)   raiz quadrada
;
; SLA                [mathfacility]
;                     Deslocamento Aritmetico a Esquerda.
;                     Desloca um registrador de 32 bits para a esquerda considerando o bit de sinal,
;                     isto e', o bit de sinal nao sera deslocado mas sim os 31 bits de mais baixa
;                     ordem.
;
; SRA                [mathfacility]
;                     Deslocamento Aritmetico a Direita
;                     Desloca um registrador de 32 bits para a direita considerando o bit de sinal,
;                     isto e', a cada bit de dado delocado, o bit de sinal sera propagado a esquerda.
;
; STACK_FSR          Liga a opcao de salvamento dos registradores FSR0, FSR1 e FSR2 no Stack de Software.
;                    Esta macro, quando emitida no codigo fonte, define um simbolo que indica para as
;                    outras macros usarem a opcao de salvamento dos registradores Indexadores no Stack 
;                    de Software.
;                    Se a opcao de salvamento for desejada, esta macro deve ser emitida antes que 
;                    qualquer outra macro aritmetica seja usada.
;
; SUBTRACT           REGA = REGA - REGB
;
; UNPACK             UNPACK orig, digits, dest               (24/12/2019)
;                    Descompacta um numero BCD de t 'digits' digitos contido no registrador 'orig', 
;                    com o formato NNN...NN, onde N pode assumir valores entre 0 e 9, no registrador 'dest'
;                    com o formato 0N0N...0N. 
;                    Para compactar, usar a macro PACK.
;                    
; ZADDW r1           [mathfacility]
;                    Zera o registrador r1 de 32 bits e carrega o seu byte menos significativo (LSB)
;                    com o valor contido no registrador W.
;
; ZADDF r1, f1       [mathfacility]
;                    Zera o registrador r1 de 32 bits e carrega o seu byte menos significativo (LSB)
;                    com o valor contido no registrador f1.
;
; ZREG  r1           [mathfacility]
;                    Zera o registrador r1 de 32 bits.
;
;
;
; -----------------------------------------------------------------------------------------------------------
; -----------------------------------------------------------------------------------------------------------
; -----------------------------------------------------------------------------------------------------------




; -----------------------------------------------------------------------------------
; Salvamento dos registradores FSR0, FSR1 e FSR2 no Stack de Software      28/11/2017
;
; As funcoes STACK_ALLFSR e UNSTACK_ALLFSR armazenam e restauram, respectivamente, os
; registradores FSR0, FSR1 e FSR2 no Stack de Software.
;
; Para que a opcao de salvamento seja usada, defina no programa, antes de incluir o
; servico math_services, o  simbolo __STACK_FSR que indicara para as outras macros 
; fazerem o salvamento dos registradores Indexadores no Stack de Software.
;
; -----------------------------------------------------------------------------------
  ifdef    __STACK_FSR
STACK_ALLFSR
           pushR FSR0H 
           pushR FSR0L
           pushR FSR1H
           pushR FSR1L
           pushR FSR2H
           pushR FSR2L
           return
UNSTACK_ALLFSR
           popR FSR2L
           popR FSR2H
           popR FSR1L
           popR FSR1H
           popR FSR0L
           popR FSR0H
           return
  endif






; ----------------------------------------------------------------------
;Compacta um numero decimal em BCD                            27/12/2019
;
;          PACK   orig, digits, dest
;
; orig ............ Registrador contendo o numero decimal no formato
;                   0n0n0n...
;
; digits .......... Numero de digitos resultantes no destino:
;                   origem = 0x0203 ------>  destino = 0x23 -----> 2 digitos
;                   origem = 0x01020304 -->  destino = 0x1234 ---> 4 digitos
;
; dest ............ Registrador que contera o resultado no formato compactado.
;                   Se a origem tiver um numero impar de digitos, esta funcao
;                   preenchera o digito MSB com zero.
;
; Nota: 
; 1 - Os registradores 'orig' e 'dest' podem residir em qualquer bancoa a memoria RAM.
; 2 - Os regigistradores indexadores FSR0, FSR1 e FSR2 poderao ser destruidos ou 
;     preservados com o uso da macro STACK_ALLFSR_MATH_REGS.
; 3 - O registrador REGC0 sera destruido.
;
; ----------------------------------------------------------------------
PACK       MACRO   orig, digits, dest
    ifndef _PACK_
#define    _PACK_  1
           goto    packExit            ; salta a subrotina na primeira vez
pack       lfsr    FSR2, MTEMP         ; aponta contador temporario
           movwf   INDF2               ; salva o numero de digitos
pack2      movwf   REGC0               ; salva numero de digitos
           movf    INDF1, w            ; pega primeiro digito na origem
           rrcf    REGC0, f            ; divide numero de digitos por 2
           btfss   STATUS, C           ; numero impar de digitos?
           goto    pack2a              ; NAO --> numero par 
           movwf   POSTINC0            ; SIM --> salva do jeito que esta
           movf    POSTINC1, w         ; somente avanca o apontador na origem
           decf    INDF2, f            ; ajusta 1 digito porque e' impar
pack2a     rrcf    INDF2, f            ; divide numero de digitos por 2 (2 digitos por byte)
           bcf     INDF2, 7            ; elimina sujeira do Carry
pack3      swapf   POSTINC1, w         ; se numero par de digitos, posiciona o MSB
           andlw   0xF0                ; zera o LSB
           iorwf   POSTINC1, w         ; insere o LSB
           movwf   POSTINC0            ; salva BCD no registrador de saida
pack4      decfsz  INDF2, f            ; terminou???
           goto    pack3               ; NAO --> proximo
           return                      ; fim
packExit
    endif
  ifdef    __STACK_FSR       ; salvamento dos registradores de trabalho
           call    STACK_ALLFSR
  endif

    if digits < 2
      error "Macro PACK - Numero minimo de digitos BCD deve ser 2"
    endif

           lfsr    FSR1, orig
           lfsr    FSR0, dest
           movlw   digits
           call    pack

  ifdef    __STACK_FSR
           call    UNSTACK_ALLFSR
  endif
           ENDM


; ----------------------------------------------------------------------
; Descompacta um numero BCD                                   20/12/2019
;
;          UNPACK   orig, digits, dest
;
; orig ............ Registrador de ate 32 bits contendo ate 8 digitos BCD (0 a 9)
;                   no formato nnnn.....
;
; digits .......... Numero de digitos na origem, exemplo: 
;                   origem = 0x23 ----> 2 digitos
;                   origem = 0x0023 --> 4 digitos
;
; dest ............ Registrador que contera o resultado no formato descompactado.
;                   Deve conter o dobro do tamanho da origem, exemplo:
;                   origem(1)=0x23 -------> destino(2)=0x0203
;                   origem(3)=0x131416 ---> destino(6)=0x010301040106
;
; Nota: 
; 1 - Os registradores 'orig' e 'dest' podem residir em qualquer bancoa a memoria RAM.
; 2 - Os regigistradores indexadores FSR0, FSR1 e FSR2 poderao ser destruidos ou 
;     preservados com o uso da macro STACK_ALLFSR_MATH_REGS.
;
; ----------------------------------------------------------------------
UNPACK     MACRO   orig, digits, dest
  ifndef _UNPACK_
#define   _UNPACK_    1
           goto    unpkExit
unpack     lfsr    FSR2, MTEMP        ; contador de digitos
           movwf   INDF2              ; numero de digitos BCD na origem
unpk2      swapf   INDF1, w
           andlw   0x0F
           movwf   POSTINC0
           movf    POSTINC1, w
           andlw   0x0F
           movwf   POSTINC0
           decfsz  INDF2, f
           goto    unpk2
           return
unpkExit
  endif

  ifdef    __STACK_FSR       ; salvamento dos registradores de trabalho
           call    STACK_ALLFSR
  endif
           lfsr    FSR1, orig 
           lfsr    FSR0, dest
           movlw   digits
           call    unpack
  ifdef    __STACK_FSR
           call    UNSTACK_ALLFSR
  endif
           ENDM 


; -----------------------------------------------------------------------------------
; CRF - Compara 2 registradores de 32 bits em qualquer lugar da RAM.       12/11/2017
;
; O todos os dois operandos podem residir em qualquer lugar da RAM.
;
; A macro STACK_ALLFSR_MATH_REGS oferece opcao de salvar no stack de software, os registradores de
; trabalho REGCx, FSR1 e FSR2. 
;
; RESTRICOES: 
; 1 - Os registradores aritmeticos REGC0 a REGC3, MCOUNT e DCOUNT nao podem ser usados
;     como operandos desta macro
;
; NOTAS:
; 1 - Os regigistradores indexadores FSR1 e FSR2 poderao ser destruidos ou
;     preservados com o uso da macro STACK_ALLFSR_MATH_REGS. 
; 2 - Os registradores aritmeticos REGC3 a REGC0 poderao ser destruidos ou
;     preservados com o uso da macro STACK_ALLFSR_MATH_REGS. 
; 3 - Os registradores aritmeticos MCOUNT e DCOUNT serao destruidos.
; 4 - O resultado da comparacao estara nos flags de STATUS Z e C.
;     Z      C   Resultado
;     0      0   Op1 < Op2   Negativo
;     0      1   Op1 > Op2   Positivo
;     1      1   Op1 = Op2   Nulo
;
; 5 - Para fazer desvios apos uma comparacao, use as macros BE, BL e BH definidas neste
;     arquivo.
; 
; -----------------------------------------------------------------------------------
CRF        MACRO   op1, op2
  ifndef __cfr_fun
#define __cfr_fun 1
           goto    crf_exit                                ; salto sobre a funcao abaixo
           #include "..\..\Services\mathcrf_services.inc"  ; funcao de comparacao de 32 bits
crf_exit
  endif
           lfsr    FSR1, op1
           lfsr    FSR2, op2
           call    crf_func
           ENDM

; -----------------------------------------------------------------------------------
; CRFI - Compara o registrador de 32 bits no primeiro operando com valor imediato.
;                                                                          12/11/2017
;
; O primeiro operando pode residir em qualquer lugar da RAM.
;
; O segundo operando deve ser um valor imediato, isto e', como 0x012286.
;
; A macro STACK_ALLFSR_MATH_REGS oferece opcao de salvar no stack de software, os registradores de
; trabalho REGCx, FSR1 e FSR2. 
;
; Apesar do registrador que esta sendo comparado ser de 32 bits, apenas seus 24 bits
; de mais baixa ordem serao comparados com o valor imediato devido a limitacao da
; linguagem assembler nao ter como carregar o valor do byte de mais alta ordem, isto
; e', so podemos carregar valores de 24 bits com UPPER(x), HIGH(x) e LOW(x).
;
; RESTRICOES: 
; 1 - Os registradores aritmeticos REGC0 a REGC3, MCOUNT e DCOUNT nao podem ser usados
;     como operandos desta macro
;
;
; NOTAS:
; 1 - Os regigistradores indexadores FSR1 e FSR2 poderao ser destruidos ou
;     preservados com o uso da macro STACK_ALLFSR_MATH_REGS. 
; 2 - Os registradores aritmeticos REGC3 a REGC0 poderao ser destruidos ou
;     preservados com o uso da macro STACK_ALLFSR_MATH_REGS. 
; 3 - Os registradores aritmeticos MCOUNT e DCOUNT serao destruidos.
; 4 - O resultado da comparacao estara nos flags de STATUS Z e C.
;     Z      C   Resultado
;     0      0   Op1 < Op2   Negativo
;     0      1   Op1 > Op2   Positivo
;     1      1   Op1 = Op2   Nulo
;
; 5 - Para fazer desvios apos uma comparacao, use as macros BE, BL e BH definidas neste
;     arquivo.
; 
; -----------------------------------------------------------------------------------
CRFI       MACRO   op1, ival
  ifndef __cfr_fun
#define __cfr_fun 1
           goto    crf_exit                                ; salto sobre a funcao abaixo
           #include "..\..\Services\mathcrf_services.inc"  ; funcao de comparacao de 32 bits
crf_exit 
 endif  

;           clrf    REGC3
  lfsr FSR0, REGC3
  clrf POSTINC0
           movlw   UPPER(ival)
;           movwf   REGC2
  movwf POSTINC0
           movlw   HIGH(ival)
;           movwf   REGC1
  movwf POSTINC0
           movlw   LOW(ival)
;           movwf   REGC0
  movwf POSTINC0
           lfsr    FSR1, op1
           lfsr    FSR2, REGC3
           call    crf_func
           ENDM

; -----------------------------------------------------------------------------------
; BE - Apos comparacao (CR Op1, Op2) desvia se Op1 == Op2                  11/11/2017
; -----------------------------------------------------------------------------------
BE     MACRO   destino
       btfsc   STATUS, Z
       goto    destino 
       ENDM

; -----------------------------------------------------------------------------------
; BE - Apos comparacao (CR Op1, Op2) desvia se Op1 > Op2                   11/11/2017
; -----------------------------------------------------------------------------------
BH     MACRO   destino             ; resultou positivo entao: Z=0 e C=1 --> destino
  ifndef _numq
       variable _numq
  endif
_numq set _numq + 1
       btfsc   STATUS, Z
       goto    _bh#v(_numq) 
       btfsc   STATUS, C
       goto    destino
_bh#v(_numq) 
       ENDM

; -----------------------------------------------------------------------------------
; BE - Apos comparacao (CR Op1, Op2) desvia se Op1 < Op2                   11/11/2017
; -----------------------------------------------------------------------------------
BL     MACRO   destino             ; resultou negativo entao: Z=0 e C=0 --> destino
  ifndef _numq
       variable _numq
  endif
_numq set _numq + 1
       btfsc   STATUS, Z
       goto    _bl#v(_numq) 
       btfss   STATUS, C
       goto    destino
_bl#v(_numq) 
       ENDM

; ---------------------------------------------------------------------------------------------
; Simbolos usados pelas macros e funcoes deste servico.                              31/12/2015
; ---------------------------------------------------------------------------------------------
REGA         equ       1           ; para uso das macros
REGB         equ       2           ; para uso das macros
REGC         equ       3           ; para uso das macros
POS          equ       0           ; para uso das macros
POSITIVE     equ       0           ; para uso das macros
NEG          equ       1           ; para uso das macros
NEGATIVE     equ       1           ; para uso das macros
RG           equ       1           ; parametro e' um registrador 
rg           equ       1           ; parametro e' um registrador 
I            equ       2           ; parametro e' um valor imediato
i            equ       2           ; parametro e' um valor imediato
;
; ---------------------------------------------------------------------------------------------
; ADD ---> REGA = REGA + REGB                                                        31/12/2015
; ---------------------------------------------------------------------------------------------
ADD          MACRO
  ifdef    __STACK_FSR       ; salvamento dos registradores de trabalho
           call    STACK_ALLFSR
  endif
             call     add

  ifdef    __STACK_FSR
           call    UNSTACK_ALLFSR
  endif
             ENDM
; ---------------------------------------------------------------------------------------------
; SUBTRACT ---> REGA = REGA - REGB                                                   31/12/2015
; ---------------------------------------------------------------------------------------------
SUBTRACT     MACRO
  ifdef    __STACK_FSR       ; salvamento dos registradores de trabalho
           call    STACK_ALLFSR
  endif
             call     subtract
  ifdef    __STACK_FSR
           call    UNSTACK_ALLFSR
  endif
             ENDM
; ---------------------------------------------------------------------------------------------
; DIVIDE ---> REGA = REGA / REGB    resto em REGC                                    31/12/2015
; ---------------------------------------------------------------------------------------------
DIVIDE       MACRO
  ifdef    __STACK_FSR       ; salvamento dos registradores de trabalho
           call    STACK_ALLFSR
  endif
             call     divide
  ifdef    __STACK_FSR
           call    UNSTACK_ALLFSR
  endif
             ENDM
; ---------------------------------------------------------------------------------------------
; MULTIPLY ---> REGA = REGA * REGB                                                   31/12/2015
; ---------------------------------------------------------------------------------------------
MULTIPLY     MACRO
  ifdef    __STACK_FSR       ; salvamento dos registradores de trabalho
           call    STACK_ALLFSR
  endif
             call     multiply
  ifdef    __STACK_FSR
           call    UNSTACK_ALLFSR
  endif
             ENDM
; ---------------------------------------------------------------------------------------------
; SQRT ---> REGA = sqrt(REGA)                                                        31/12/2015
; ---------------------------------------------------------------------------------------------
SQRT         MACRO
  ifdef    __STACK_FSR       ; salvamento dos registradores de trabalho
           call    STACK_ALLFSR
  endif
             call     sqrt
  ifdef    __STACK_FSR
           call    UNSTACK_ALLFSR
  endif
             ENDM
; ---------------------------------------------------------------------------------------------
; BIN2DEC --> binario em REGA para decimal em DIGIT1 a DIGIT10                       31/12/2015
; ---------------------------------------------------------------------------------------------
BIN2DEC      MACRO
  ifdef    __STACK_FSR       ; salvamento dos registradores de trabalho
           call    STACK_ALLFSR
  endif
             call     bin2dec
  ifdef    __STACK_FSR
           call    UNSTACK_ALLFSR
  endif
             ENDM
; ---------------------------------------------------------------------------------------------
; DEC2BIN --> Converte o decimal em DIGIT1 a DIGITn para binario                     31/12/2015
;             com resultado em REGA.
;             Se o parametro 'len' for um valor imediato, o parametro 'lenopt' deve ser 'I' ou
;             'i', sem aspas.
;             Se o parametro 'len' for um registrador, o parametro 'lenopt' deve ser 'RG' ou
;             'rg', sem aspas.
;             O parametro 'dim' deve ser NEG se o valor for negativo ou POS se positivo.
;
; ---------------------------------------------------------------------------------------------
DEC2BIN     MACRO   len, lenopt, dim
  ifdef    __STACK_FSR       ; salvamento dos registradores de trabalho
           call    STACK_ALLFSR
  endif
            lfsr    FSR1, DSIGN
            clrf    INDF1            ; assume que o numero e' positivo

       if dim==NEG
            bsf     INDF1, 0         ; o numero e' negativo
       endif

       if (lenopt==rg || lenopt==RG)
            movf    len, w
       endif


       if (lenopt==i || lenopt==I)
            movlw   len
       endif


            call    dec2bin

  ifdef    __STACK_FSR
           call    UNSTACK_ALLFSR
  endif
            ENDM
; ---------------------------------------------------------------------------------------------
; NEGATE --> REGx = -REGx    sendo x=A ou x=B                                        31/12/2015
; ---------------------------------------------------------------------------------------------
NEGATE       MACRO    regmath
  ifdef    __STACK_FSR       ; salvamento dos registradores de trabalho
           call    STACK_ALLFSR
  endif
      if regmath==REGA
         call   negatea
      endif
      if regmath==REGB
         call   negateb
      endif
  ifdef    __STACK_FSR
           call    UNSTACK_ALLFSR
  endif
             ENDM
; ---------------------------------------------------------------------------------------------
; CLRA --> REGA = 0                                                                  31/12/2015
; ---------------------------------------------------------------------------------------------
CLRA         MACRO
  ifdef    __STACK_FSR       ; salvamento dos registradores de trabalho
           call    STACK_ALLFSR
  endif
             call   clra
  ifdef    __STACK_FSR
           call    UNSTACK_ALLFSR
  endif
             ENDM
; ---------------------------------------------------------------------------------------------
; CLRB --> REGB = 0                                                                  31/12/2015
; ---------------------------------------------------------------------------------------------
CLRB         MACRO
  ifdef    __STACK_FSR       ; salvamento dos registradores de trabalho
           call    STACK_ALLFSR
  endif
             call   clrb
  ifdef    __STACK_FSR
           call    UNSTACK_ALLFSR
  endif
             ENDM
; ---------------------------------------------------------------------------------------------
; CLRDIG --> DIGIT1 a DIGIT10 = 0                                                    31/12/2015 
; ---------------------------------------------------------------------------------------------
CLRDIG       MACRO
  ifdef    __STACK_FSR       ; salvamento dos registradores de trabalho
           call    STACK_ALLFSR
  endif
             call   clrdig
  ifdef    __STACK_FSR
           call    UNSTACK_ALLFSR
  endif
             ENDM
; ---------------------------------------------------------------------------------------------
; MATHLD - Carrega um registrador aritmetico (32bits) com um valor imediato de 24 bits.
;          O primeiro operando e' o registrador aritmetico.
;          O segundo operando e' o valor imediato que devera ser sempre positivo entre 0 (zero)
;          e 16.777.215 (0xFFFFFF) no maximo, inclusive.
;                                                                                    31/12/2015
; ---------------------------------------------------------------------------------------------
MATHLD        MACRO     regmath, valor

  ifdef    __STACK_FSR             
             call      STACK_ALLFSR         ; salva indexadores
  endif

  if regmath==REGA
             lfsr      FSR1, REGA3          ; indexa REGA
  endif

  if regmath==REGB
             lfsr      FSR1, REGB3          ; indexa REGB
  endif

  if regmath==REGC
             lfsr      FSR1, REGC3          ; indexa REGC
  endif

  ; 22/11/2017 *** Aceitar valor negativo do segundo operando.
  ; Se o valor imediato e' negativo, sera forcado o sinal no MSB no registrador aritmetico.
  if valor < 0
             movlw     0xFF                 ; sinal negativo
             movwf     POSTINC1             ; assume negativo
  else
             clrf      POSTINC1             ; assume positivo
  endif
             movlw     upper(valor)         ; UPPER
             movwf     POSTINC1
             movlw     high(valor)          ; HIGH
             movwf     POSTINC1
             movlw     low(valor)           ; LOW
             movwf     POSTINC1

  ifdef    __STACK_FSR
             call      UNSTACK_ALLFSR       ; restaura indexadores
  endif
             ENDM
; ---------------------------------------------------------------------------------------------
; MATHST - Armazena um registrador aritmetico num registrador qualquer  (32bits). 
;          O primeiro operando e' deve ser um registrador aritmetico como REGA, REGB ou REGC
;          cujo valor sera movido para o segundo operando.
;          O segundo operando e' o registrador que recebera o valor do registrador aritmetico.
;                                                                                    31/12/2015
; ---------------------------------------------------------------------------------------------
MATHST        MACRO     regmath, regx

  ifdef    __STACK_FSR       ; salvamento dos registradores de trabalho
           call    STACK_ALLFSR
  endif
  if regmath == REGA
             lfsr      FSR1, REGA3
  endif

  if regmath == REGB
             lfsr      FSR1, REGB3
  endif

  if regmath == REGC
             lfsr      FSR1, REGC3
  endif        
             movff     POSTINC1, regx+.0
             movff     POSTINC1, regx+.1
             movff     POSTINC1, regx+.2
             movff     POSTINC1, regx+.3

  ifdef    __STACK_FSR
           call    UNSTACK_ALLFSR
  endif
             ENDM
; ---------------------------------------------------------------------------------------------
; MATHLR - Armazena um registrador qualquer (32bits) do segundo operando num registrador 
;          aritmetico do primeiro operando.
;          O primeiro operando deve ser um registrador aritmetico como REGA, REGB ou REGC
;          que recebera o valor do segundo operando.
;          O segundo operando e' o registrador cujo valor sera movido para o registrador
;          aritmetico.
;                                                                                    31/12/2015
; ---------------------------------------------------------------------------------------------
MATHLR       MACRO     regmath, regx

  ifdef    __STACK_FSR       ; salvamento dos registradores de trabalho
           call    STACK_ALLFSR
  endif
  if regmath == REGA
             lfsr      FSR1, REGA3
  endif

  if regmath == REGB
             lfsr      FSR1, REGB3
  endif

  if regmath == REGC
             lfsr      FSR1, REGC3
  endif        
             movff     regx+.0, POSTINC1
             movff     regx+.1, POSTINC1
             movff     regx+.2, POSTINC1
             movff     regx+.3, POSTINC1

  ifdef    __STACK_FSR
           call    UNSTACK_ALLFSR
  endif
             ENDM
; ---------------------------------------------------------------------------------------------
; MATHASCII - Insere, nos nible MSB dos registradores DIGIT1 a DIGIT10, a configuracao de bits 
;             b'0011' apos uma conversao de binario para decimal (bin2dec) a fim de transformar
;             os valores desses registradores em formato ASCII.
;             Todos os registradores DIGITx serao alterados.
;
; EXEMPLO:
;     MATHLD  REGA, .21343   ; carrega um valor em REGA
;     call    bin2dec        ; converte a decimal em DIGIT1 a DIGIT10 (00000000000201030403)
;     MATHASCII              ; transforma DIGIT1 a DIGIT10 em ASCII   (30303030303231333433)
;
;                                                                                    31/12/2015
; ---------------------------------------------------------------------------------------------
MATHASCII    MACRO 
  ifdef    __STACK_FSR       ; salvamento dos registradores de trabalho
           call    STACK_ALLFSR
  endif
  ifndef __asciize_indicator
#define  __asciize_indicator 1
             goto     _asciize_03
_asciize_01  lfsr     FSR1, DIGIT1
             lfsr     FSR2, MCOUNT
             movlw    .10
             movwf    INDF2
             movlw    0x30
_asciize_02  iorwf    POSTINC1, f
             decfsz   INDF2, f
             goto     _asciize_02
             return
_asciize_03
  endif
  call     _asciize_01

  ifdef    __STACK_FSR
           call    UNSTACK_ALLFSR
  endif
             ENDM

; ---------------------------------------------------------------------------
; 32 BIT SIGNED SUTRACT                                            31/12/2015
; REGA - REGB -> REGA
; Return carry set if overflow
;
; ---------------------------------------------------------------------------
    GLOBAL  subtract
subtract
	call	negateb     ; Negate REGB
    btfsc   STATUS, C   ;
	return              ; Overflow
; ---------------------------------------------------------------------------
; 32 BIT SIGNED ADD                                                31/12/2015
; REGA + REGB -> REGA
; Return carry set if overflow
; ---------------------------------------------------------------------------
    GLOBAL  add
add	lfsr    FSR1, REGA3
    movf    INDF1, w        ; movf	REGA3,w		;Compare signs
    lfsr    FSR1, REGB3
	xorwf   INDF1, w        ; xorwf	REGB3,w
    lfsr    FSR1, MTEMP
	movwf   INDF1           ; movwf	MTEMP
	call	addba           ; Add REGB to REGA
	bcf     STATUS,C        ; Check signs
    lfsr    FSR1, REGB3
	movf    INDF1, w        ; movf	REGB3,w		;If signs are same
    lfsr    FSR1, REGA3
	xorwf   INDF1, w        ; xorwf	REGA3,w		;so must result sign
    lfsr    FSR1, MTEMP
	btfss   INDF1, 7        ; btfss	MTEMP,7		;else overflow
	addlw	0x80            ; to set Z and C in PIC16F not works in PIC18F
	return
; ---------------------------------------------------------------------------
; 32 BIT SIGNED MULTIPLY                                           31/12/2015
; REGA * REGB -> REGA
; Return carry set if overflow
;
; So funciona quando (A>0 e B>0) ou  (A<0 e B>0) 
; Nao funciona quando B<0
; ---------------------------------------------------------------------------
    GLOBAL multiply
multiply
    lfsr    FSR1, MTEMP
	clrf    INDF1           ; clrf	MTEMP		;Reset sign flag
	call	absa            ; Make REGA positive
	btfss   STATUS,C
	call	absb            ; Make REGB positive
	btfsc   STATUS,C
	return                  ; Overflow
	call	movac           ; Move REGA to REGC
	call	clra            ; Clear product
	movlw	D'31'           ; Loop counter
    lfsr    FSR1, MCOUNT
	movwf   INDF1           ; movwf	MCOUNT
muloop
	call	slac            ; Shift left product and multiplicand
    lfsr    FSR1, REGC3
	rlcf    INDF1, w        ; rlcf	REGC3,w         ; Test MSB of multiplicand
	btfsc   STATUS,C        ; If multiplicand bit is a 1 then
	call	addba           ; add multiplier to product
	btfss   STATUS,C        ; Check for overflow
    lfsr    FSR1, REGA3
	rlcf    INDF1, w        ; rlcf	REGA3,w
	btfsc    STATUS,C
	return
    lfsr    FSR1, MCOUNT
	decfsz  INDF1, f        ; decfsz	MCOUNT,f	;Next
	goto	muloop
    lfsr    FSR1, MTEMP
	btfsc   INDF1, 0        ; btfsc	MTEMP,0		;Check result sign
	call	negatea         ; Negative
	return
; ---------------------------------------------------------------------------
; 32 BIT SIGNED DIVIDE                                             31/12/2015
; REGA / REGB -> REGA
; Remainder in REGC
; Return carry set if overflow or division by zero
;
; So funciona quando (A>0 e B>0) ou  (A<0 e B>0) 
; Nao funciona quando B<0
; ---------------------------------------------------------------------------
    GLOBAL  divide
divide
    lfsr    FSR1, MTEMP
	clrf    INDF1          ; clrf	MTEMP		;Reset sign flag
    lfsr    FSR1, REGB0
	movf    POSTDEC1, w    ; movf	REGB0,w		;Trap division by zero
	iorwf   POSTDEC1, w    ; iorwf	REGB1,w
	iorwf	POSTDEC1, w    ; iorwf  REGB2,w
	iorwf	POSTDEC1, w    ; iorwf  REGB3,w
	sublw	0
	btfss   STATUS,C
	call	absa           ; Make dividend (REGA) positive
	btfss   STATUS,C
	call	absb           ; Make divisor (REGB) positive
	btfsc   STATUS,C
	return                 ; Overflow
	call    clrc           ; clrf	REGC0		;Clear remainder
                           ; clrf	REGC1
                           ; clrf	REGC2
                           ; clrf	REGC3
	call	slac           ; Purge sign bit
	movlw	D'31'          ; Loop counter
    lfsr    FSR1, MCOUNT
	movwf   INDF1          ; movwf	MCOUNT
dvloop
	call	slac           ; Shift dividend (REGA) msb into remainder (REGC)
    lfsr    FSR1, REGB3
	movf    INDF1, w       ; movf	REGB3,w	;Test if remainder (REGC) >= divisor (REGB)
    lfsr    FSR1, REGC3
	subwf   INDF1, w       ; subwf	REGC3,w
	btfss   STATUS,Z
	goto	dtstgt
    lfsr    FSR1, REGB2
	movf    INDF1, w       ; movf	REGB2,w
    lfsr    FSR1, REGC2
	subwf   INDF1, w       ; subwf	REGC2,w
	btfss   STATUS,Z
	goto	dtstgt
    lfsr    FSR1, REGB1
	movf    INDF1, w       ; movf	REGB1,w
    lfsr    FSR1, REGC1
	subwf   INDF1, w       ; subwf	REGC1,w
	btfss   STATUS,Z
	goto	dtstgt
    lfsr    FSR1, REGB0
	movf    INDF1, w       ; movf	REGB0,w
    lfsr    FSR1, REGC0
	subwf   INDF1, w       ; subwf	REGC0,w
dtstgt
	btfss   STATUS,C       ; Carry set if remainder >= divisor
	goto	dremlt
    lfsr    FSR1, REGB0
	movf    INDF1, w       ; movf	REGB0,w		;Subtract divisor (REGB) from remainder (REGC)
    lfsr    FSR1, REGC0
	subwf   INDF1, f       ; subwf	REGC0,f
    lfsr    FSR1, REGB1
    lfsr    FSR2, REGC1
	movf    INDF1, w       ; movf	REGB1,w
	btfss   STATUS,C
	incfsz  INDF1, w       ; incfsz	REGB1,w
	subwf   INDF2, f       ; subwf	REGC1,f
    lfsr    FSR1, REGB2
    lfsr    FSR2, REGC2
	movf    INDF1, w       ; movf	REGB2,w
	btfss   STATUS,C
	incfsz  INDF1, w       ; incfsz	REGB2,w
	subwf   INDF2, f       ; subwf	REGC2,f
    lfsr    FSR1, REGB3
    lfsr    FSR2, REGC3
	movf    INDF1, w       ; movf	REGB3,w
	btfss   STATUS,C
	incfsz  INDF1, w       ; incfsz	REGB3,w
	subwf   INDF2, f       ; subwf	REGC3,f
	bcf     STATUS,C
    lfsr    FSR1, REGA0
	bsf     INDF1, 0       ; bsf     REGA0,0		;Set quotient bit
dremlt
    lfsr    FSR1, MCOUNT
	decfsz  INDF1, f       ; decfsz	MCOUNT,f	;Next
	goto	dvloop
    lfsr    FSR1, MTEMP
	btfsc   INDF1, 0       ; btfsc	MTEMP,0		;Check result sign
	call	negatea        ; Negative
	return
; ---------------------------------------------------------------------------
; ROUND RESULT OF DIVISION TO NEAREST INTEGER                      31/12/2015
; ---------------------------------------------------------------------------
round
    lfsr    FSR1, MTEMP
	clrf    INDF1             ; clrf	MTEMP		;Reset sign flag
	call	absa              ; Make positive
	bcf     STATUS,C
	call	slc               ; Multiply remainder by 2
    lfsr    FSR1, REGB3
	movf    INDF1, w          ; movf	REGB3,w           ; Test if remainder (REGC) >= divisor (REGB)
    lfsr    FSR1, REGC3
	subwf   INDF1, w          ; subwf	REGC3,w
	btfss    STATUS,Z
	goto	rtstgt
    lfsr    FSR1, REGB2
	movf    INDF1, w          ; movf	REGB2,w
    lfsr    FSR1, REGC2
	subwf   INDF1, w          ; subwf	REGC2,w
	btfss    STATUS,Z
	goto	dtstgt
    lfsr    FSR1, REGB1
	movf    INDF1, w          ; movf	REGB1,w
    lfsr    FSR1, REGC1
	subwf   INDF1, w          ; subwf	REGC1,w
	btfss    STATUS,Z
	goto	rtstgt
    lfsr    FSR1, REGB0
	movf    INDF1, w          ; movf	REGB0,w
    lfsr    FSR1, REGC0
	subwf   INDF1, w          ; subwf	REGC0,w
rtstgt
	btfss    STATUS,C         ; Carry set if remainder >= divisor
	goto	rremlt
    lfsr    FSR1, REGA0
	incfsz  POSTDEC1, f       ; incfsz	REGA0,f		;Add 1 to quotient
	goto	rremlt
	incfsz  POSTDEC1, f       ; incfsz	REGA1,f
	goto	rremlt
	incfsz  POSTDEC1, f       ; incfsz	REGA2,f
	goto	rremlt
	incf    POSTDEC1, f       ; incf	REGA3,f
	btfsc    STATUS,Z
	return                    ; Overflow,return carry set
rremlt
    lfsr    FSR1, MTEMP
	btfsc   INDF1, 0          ; btfsc	MTEMP,0           ; Restore sign
	call	negatea
	return
; ---------------------------------------------------------------------------
; 32 BIT SQUARE ROOT                                               31/12/2015
; sqrt(REGA) -> REGA
; Return carry set if negative
; ---------------------------------------------------------------------------
    GLOBAL  sqrt
sqrt	
    lfsr    FSR1, REGA3
    rlcf    INDF1, w          ; rlcf	REGA3,w		;Trap negative values
	btfsc   STATUS,C
	return

	call	movac             ; Move REGA to REGC
	call	clrba             ; Clear remainder (REGB) and root (REGA)
    lfsr    FSR1, MCOUNT
	movlw	D'16'             ; Loop counter
	movwf   INDF1             ; movwf	MCOUNT

sqloop
    lfsr    FSR1, REGC0
	rlcf    POSTDEC1, f       ; rlcf	REGC0,f		;Shift two msb's
	rlcf    POSTDEC1, f       ; rlcf	REGC1,f		;into remainder
	rlcf    POSTDEC1, f       ; rlcf	REGC2,f
	rlcf    POSTDEC1, f       ; rlcf	REGC3,f
    lfsr    FSR1, REGB0
	rlcf    POSTDEC1, f       ; rlcf	REGB0,f
	rlcf    POSTDEC1, f       ; rlcf	REGB1,f
	rlcf    POSTDEC1, f       ; rlcf	REGB2,f
    lfsr    FSR1, REGC0
	rlcf    POSTDEC1, f       ; rlcf	REGC0,f
	rlcf    POSTDEC1, f       ; rlcf	REGC1,f
	rlcf    POSTDEC1, f       ; rlcf	REGC2,f
	rlcf    POSTDEC1, f       ; rlcf	REGC3,f
    lfsr    FSR1, REGB0
	rlcf    POSTDEC1, f       ; rlcf	REGB0,f
	rlcf    POSTDEC1, f       ; rlcf	REGB1,f
	rlcf    POSTDEC1, f       ; rlcf	REGB2,f

	bsf     STATUS,C          ; Add 1 to root
    lfsr    FSR1, REGA0
	rlcf    POSTDEC1, f       ; rlcf	REGA0,f		;Align root
	rlcf    POSTDEC1, f       ; rlcf	REGA1,f
	rlcf    POSTDEC1, f       ; rlcf	REGA2,f

    lfsr    FSR1, REGA2
	movf    INDF1, w          ; movf	REGA2,w		;Test if remdr (REGB) >= root (REGA)
    lfsr    FSR1, REGB2
	subwf   INDF1, w          ; subwf	REGB2,w
	btfss    STATUS,Z
	goto	ststgt
    lfsr    FSR1, REGA1
	movf    INDF1, w          ; movf	REGA1,w
    lfsr    FSR1, REGB1
	subwf   INDF1, w          ; subwf	REGB1,w
	btfss    STATUS,Z
	goto	ststgt
    lfsr    FSR1, REGA0
	movf    INDF1, w          ; movf	REGA0,w
    lfsr    FSR1, REGB0
	subwf   INDF1, w          ; subwf	REGB0,w
ststgt
	btfss    STATUS,C         ; Carry set if remdr >= root
	goto	sremlt
    lfsr    FSR1, REGA0
	movf    INDF1, w          ; movf	REGA0,w  ; Subtract root (REGA) from remdr (REGB)
    lfsr    FSR1, REGB0
	subwf   INDF1, f          ; subwf	REGB0,f
    lfsr    FSR1, REGA1
    lfsr    FSR2, REGB1
	movf    INDF1, w          ; movf	REGA1,w
	btfss    STATUS,C
	incfsz  INDF1, w          ; incfsz	REGA1,w
	subwf   INDF2, f          ; subwf	REGB1,f
    lfsr    FSR1, REGA2
    lfsr    FSR2, REGB2
	movf    INDF1, w          ; movf	REGA2,w
	btfss    STATUS,C
	incfsz   INDF1, w         ; incfsz	REGA2,w
	subwf    INDF2, f         ; subwf	REGB2,f
    lfsr     FSR1, REGA0
	bsf      INDF1, 1         ; bsf	REGA0,1		;Set current root bit

sremlt
    lfsr     FSR1, REGA0
	bcf      INDF1, 0         ; bcf	REGA0,0		;Clear test bit
    lfsr     FSR1, MCOUNT
	decfsz   INDF1, f         ; decfsz	MCOUNT,f	;Next
	goto     sqloop

	bcf     STATUS,C
    lfsr    FSR1, REGA2
	rrcf    POSTINC1, f       ; rrcf	REGA2,f		;Adjust root alignment
	rrcf    POSTINC1, f       ; rrcf	REGA1,f
	rrcf    POSTINC1, f       ; rrcf	REGA0,f
	return
; ---------------------------------------------------------------------------
; 32 BIT SIGNED BINARY TO DECIMAL                                  31/12/2015
; REGA -> DIGITS 1 (MSD) TO 10 (LSD) & DSIGN
; DSIGN = 0 if REGA is positive, 1 if negative
; Return carry set if overflow
; Uses FSR register
; ---------------------------------------------------------------------------
    GLOBAL bin2dec
bin2dec
    lfsr    FSR1, MTEMP
    clrf    INDF1              ; clrf	MTEMP		;Reset sign flag
	call	absa               ; Make REGA positive
	btfsc   STATUS,C
	return                     ; Overflow
	call    clrdig             ; Clear all digits
	movlw	D'32'              ; Loop counter
    lfsr    FSR1, MCOUNT
    movwf   INDF1              ; movwf	MCOUNT

b2dloop	
    lfsr    FSR1, REGA0
    rlcf    POSTDEC1, f        ; rlcf	REGA0,f	 ;Shift msb into carry 
    rlcf    POSTDEC1, f        ; rlcf	REGA1,f
    rlcf    POSTDEC1, f        ; rlcf	REGA2,f
    rlcf    POSTDEC1, f        ; rlcf	REGA3,f
	lfsr	FSR0,DIGIT10       ; Pointer to digits
    lfsr    FSR1, DCOUNT
	movlw	D'10'              ; 10 digits to do
	movwf   INDF1              ; movwf	DCOUNT

adjlp	
    rlcf	INDF0,f            ; Shift digit and carry 1 bit left

	movlw	D'10'
	subwf	INDF0,w	           ; Check and adjust for decimal overflow
	btfsc   STATUS,C
	movwf	INDF0
    movf    POSTDEC0,w         ; falso move para decrementar FSR0
    lfsr    FSR1, DCOUNT
	decfsz  INDF1, f           ; decfsz	DCOUNT,f
	goto	adjlp
    lfsr    FSR1, MCOUNT
	decfsz  INDF1, f           ; decfsz	MCOUNT,f	;Next bit
	goto	b2dloop
    lfsr    FSR1, MTEMP
;	btfsc   INDF1, 0           ; btfsc	MTEMP,0		;Check sign
    btfss   INDF1, 0
    goto    adjlp1
	bsf     DSIGN,0            ; Negative
adjlp1
	bcf     STATUS,C
	return
; ---------------------------------------------------------------------------
; 32 BIT SIGNED DECIMAL TO BINARY                                  31/12/2015
; Decimal DIGIT1 thro DIGIT(X) & DSIGN -> REGA
; Set DSIGN = 0 for positive, DSIGN = 1 for negative values
; Most significant digit in DIGIT1
; Enter this routine with digit count in w register
; Return carry set if overflow
; Uses FSR register
; ---------------------------------------------------------------------------
    GLOBAL  dec2bin
dec2bin
    lfsr    FSR1, MTEMP
	movwf   INDF1         ; movwf	MTEMP		;Save digit count
    lfsr    FSR1, MCOUNT
	movlw	D'32'         ; Outer bit loop counter
	movwf   INDF1         ; movwf	MCOUNT

d2blp1
	lfsr	FSR0,DSIGN    ; Set up pointer to MSD, isto e, DIGIT1-1
	movff   MTEMP, DCOUNT ; movf	MTEMP,w		;Inner digit loop counter
	                      ; movwf	DCOUNT
	movlw	D'10'
	bcf     STATUS,C      ; Bring in '0' bit into MSD
    movff   STATUS,REGC3  ; movff   STATUS,SVStatus ; ALTERACAO EM 21/12/2017
d2blp2	
    incf	FSR0L,f
    movff   REGC3,STATUS  ; movff   SVStatus,STATUS ; ALTERACAO EM 21/12/2017
	btfsc   STATUS,C
	addwf	INDF0,f       ; Add 10 if '1' bit from prev digit
	rrcf	INDF0,f       ; Shift out LSB of digit
    movff   STATUS,REGC3  ; movff   STATUS,SVStatus ; ALTERACAO EM 21/12/2017
    lfsr    FSR1, DCOUNT
	decfsz  INDF1, f      ; decfsz	DCOUNT,f	;Next L.S. Digit
	goto	d2blp2
    lfsr    FSR1, REGA3
	rrcf    POSTINC1, f   ; rrcf	REGA3,f		;Shift in carry from digits
	rrcf    POSTINC1, f   ; rrcf	REGA2,f
	rrcf    POSTINC1, f   ; rrcf	REGA1,f
	rrcf    POSTINC1, f   ; rrcf	REGA0,f
    lfsr    FSR1, MCOUNT
	decfsz  INDF1, f      ; decfsz	MCOUNT,f	;Next bit
	goto	d2blp1
    lfsr    FSR1, REGA3
 	movf	INDF0,w       ; Check for overflow
	addlw	0xFF
	btfss   STATUS,C
	rlcf    INDF1, w      ; rlcf	REGA3,w
	btfsc   STATUS,C
	return
    lfsr    FSR1, DSIGN
	btfsc   INDF1, 0      ; btfsc	DSIGN,0		;Check result sign
	call	negatea       ; Negative
	return

; ---------------------------------------------------------------------------
; Add REGB to REGA (Unsigned)                                      31/12/2015
; Used by add, multiply,
; ---------------------------------------------------------------------------
addba
    lfsr    FSR1, REGB0
	movf    INDF1, w          ; movf	REGB0,w		;Add lo byte
    lfsr    FSR1, REGA0
	addwf   INDF1, f          ; addwf	REGA0,f
    lfsr    FSR1, REGB1
    lfsr    FSR2, REGA1
	movf    INDF1, w          ; movf	REGB1,w		;Add mid-lo byte
	btfsc   STATUS,C          ; No carry_in, so just add
	incfsz  INDF1, w          ; incfsz	REGB1,w		;Add carry_in to REGB
	addwf   INDF2, f          ; addwf	REGA1,f		;Add and propagate carry_out
    lfsr    FSR1, REGB2
    lfsr    FSR2, REGA2
	movf    INDF1, w          ; movf	REGB2,w		;Add mid-hi byte
	btfsc   STATUS,C
	incfsz  INDF1, w          ; incfsz	REGB2,w
	addwf   INDF2, f          ; addwf	REGA2,f
    lfsr    FSR1, REGB3
    lfsr    FSR2, REGA3
	movf    INDF1, w          ; movf	REGB3,w		;Add hi byte
	btfsc   STATUS,C
	incfsz  INDF1, w          ; incfsz	REGB3,w
	addwf   INDF2, f          ; addwf	REGA3,f
	return
; ---------------------------------------------------------------------------
; Move REGA to REGC                                                31/12/2015
; Used by multiply, sqrt
; ---------------------------------------------------------------------------
movac
	movff   REGA0, REGC0      ; movf	REGA0,w
	                          ; movwf	REGC0
	movff   REGA1, REGC1      ; movf	REGA1,w
	                          ; movwf	REGC1
	movff   REGA2, REGC2      ; movf	REGA2,w
                              ; movwf	REGC2
	movff   REGA3, REGC3      ; movf	REGA3,w
	                          ; movwf	REGC3
	return
; ---------------------------------------------------------------------------
; Clear REGB and REGA                                              31/12/2015
; Used by sqrt
; ---------------------------------------------------------------------------
clrba   lfsr    FSR1, REGB0
        clrf	POSTDEC1        ; REGB0
    	clrf	POSTDEC1        ; REGB1
	    clrf	POSTDEC1        ; REGB2
	    clrf	POSTDEC1        ; REGB3
; ---------------------------------------------------------------------------
; Limpa REGA0 a REGA3                                              31/12/2015
; ---------------------------------------------------------------------------
        GLOBAL  clra
clra    lfsr    FSR1, REGA0
        clrf	POSTDEC1        ; REGA0
    	clrf	POSTDEC1        ; REGA1
	    clrf	POSTDEC1        ; REGA2
	    clrf	POSTDEC1        ; REGA3
	    return

; ---------------------------------------------------------------------------
; Limpa REGB0 a REGB3                                              31/12/2015
; ---------------------------------------------------------------------------
        GLOBAL  clrb
clrb    lfsr    FSR1, REGB0
        clrf	POSTDEC1        ; REGB0
    	clrf	POSTDEC1        ; REGB1
	    clrf	POSTDEC1        ; REGB2
	    clrf	POSTDEC1        ; REGB3
	    return
; ---------------------------------------------------------------------------
; Limpa REGB0 a REGB3                                              31/12/2015
; ---------------------------------------------------------------------------
        GLOBAL  clrc
clrc    lfsr    FSR1, REGC0
        clrf	POSTDEC1        ; REGC0
    	clrf	POSTDEC1        ; REGC1
	    clrf	POSTDEC1        ; REGC2
	    clrf	POSTDEC1        ; REGC3
	    return
; ---------------------------------------------------------------------------
; Check sign of REGA and convert negative to positive              31/12/2015
; Used by multiply, divide, bin2dec, round
; ---------------------------------------------------------------------------
    GLOBAL  absa
absa	
    lfsr    FSR1, REGA3
    rlcf    INDF1, w           ; REGA3,w
	btfss   STATUS,C
	return                     ; Positive

; ------------------------------------------------------------------
; Negate REGA                                             31/12/2015
; Used by absa, multiply, divide, bin2dec, dec2bin, round que entram
; nesta rotina com MTEMP=0
; ------------------------------------------------------------------
    GLOBAL  negatea
negatea
    lfsr    FSR1, REGA3
	movf	INDF1, w         ; REGA3,w  Save sign in w
	andlw	0x80             
    lfsr    FSR1, REGA3      ;        SE REGA = -2        1111.1111 1111.1111 1111.1111 1111.1110  
	comf	POSTINC1, f      ; REGA3,f --> 2's complement 0000.0000 1111.1111 1111.1111 1111.1110   
	comf	POSTINC1, f      ; REGA2,f                    0000.0000 0000.0000 1111.1111 1111.1110 
	comf	POSTINC1, f      ; REGA1,f                    0000.0000 0000.0000 0000.0000 1111.1110 
	comf	INDF1, f         ; REGA0,f                    0000.0000 0000.0000 0000.0000 0000.0001 
	incfsz	POSTDEC1, f      ; REGA0 = REGA0 + 1          0000.0000 0000.0000 0000.0000 0000.0010 
	goto	nega1
	incfsz	POSTDEC1, f      ; REGA1 = REGA1 + 1 
	goto	nega1
	incfsz	POSTDEC1, f      ; REGA2 = REGA2 + 1
	goto	nega1
	incf	POSTDEC1, f      ; REGA3 = REGA3 + 1
nega1
    lfsr    FSR1, MTEMP
	incf	INDF1, f         ; MTEMP,f   flip sign flag   MTEMP = 0000.0001
    lfsr    FSR1, REGA3
	addwf	INDF1, w         ; REGA3,w   Return carry set if -2147483648
	return
; ------------------------------------------------------------------
; Check sign of REGB and convert negative to positive
; Used by multiply, divide
;                                                         31/12/2015
; ------------------------------------------------------------------
    GLOBAL  absb
absb	
    lfsr    FSR1, REGB3
    rlcf	INDF1, w        ; REGB3,w (em 16/01/2018 era INDF1, f)
	btfss   STATUS,C
	return                  ; Positive

; ------------------------------------------------------------------
; Negate REGB                                             31/12/2015
; Used by absb, subtract, multiply, divide
; ------------------------------------------------------------------
    GLOBAL  negateb
negateb
    lfsr    FSR1, REGB3
	movf	INDF1,w	        ;Save sign in w
	andlw	0x80
    lfsr    FSR1, REGB3
	comf	POSTINC1,f		; REGB3 --> 2's complement
	comf	POSTINC1,f		; REGB2 --> 2's complement
	comf	POSTINC1,f		; REGB1 --> 2's complement
	comf	INDF1,f         ; REGB0 --> 2's complement
	incfsz	POSTDEC1,f      ; REGB0 = REGB0 + 1
	goto	negb1
	incfsz	POSTDEC1,f      ; REGB1 = REGB1 + 1
	goto	negb1
	incfsz	POSTDEC1,f      ; REGB2 = REGB2 + 1
	goto	negb1
	incf	POSTDEC1,f      ; REGB3 = REGB3 + 1
negb1
    lfsr    FSR1, MTEMP
	incf	INDF1, f        ;flip sign flag
    lfsr    FSR1, REGB3
	addwf	INDF1, w        ;Return carry set if -2147483648
	return
; ------------------------------------------------------------------
; Shift left REGA and REGC                                31/12/2015
; Used by multiply, divide, round
; ------------------------------------------------------------------
slac
    lfsr    FSR1, REGA0
    rlcf	POSTDEC1, f      ;  REGA0,f
	rlcf	POSTDEC1, f      ; 	REGA1,f
	rlcf	POSTDEC1, f      ; 	REGA2,f
	rlcf	POSTDEC1, f      ; 	REGA3,f
    lfsr    FSR1, REGC0
slc	rlcf	POSTDEC1, f      ; REGC0,f
	rlcf	POSTDEC1, f      ; REGC1,f
	rlcf	POSTDEC1, f      ; REGC2,f
	rlcf	POSTDEC1, f      ; REGC3,f
	return
; ---------------------------------------------------------------------------
; Limpa DIGIT1 a DIGIT10                                           31/12/2015
; ---------------------------------------------------------------------------
    GLOBAL  clrdig
clrdig
    lfsr    FSR1, DSIGN
    clrf    INDF1           ; clrf   DSIGN
    lfsr    FSR1, DIGIT1
    clrf    POSTINC1        ; clrf   DIGIT1
    clrf    POSTINC1        ; clrf   DIGIT2
    clrf    POSTINC1        ; clrf   DIGIT3
    clrf    POSTINC1        ; clrf   DIGIT4
    clrf    POSTINC1        ; clrf   DIGIT5
    clrf    POSTINC1        ; clrf   DIGIT6
    clrf    POSTINC1        ; clrf   DIGIT7
    clrf    POSTINC1        ; clrf   DIGIT8
    clrf    POSTINC1        ; clrf   DIGIT9
    clrf    POSTINC1        ; clrf   DIGIT10
	return
; ---------------------------------------------------------------------------
; ---------------------------------------------------------------------------
; ---------------------------------------------------------------------------
; ---------------------------------------------------------------------------
; ---------------------------------------------------------------------------
; ---------------------------------------------------------------------------
; ---------------------------------------------------------------------------


; -----------------------------------------------------------------------------------
; -----------------------------------------------------------------------------------
; -----------------------------------------------------------------------------------
    #include "..\..\Services\mathfacility_services.inc"


   endif    ; ifndef __MATH_SERVICES




