; ---------------------------------------------------------------------------------------------
; CONTROLANDO UM SERVO MOTOR TIPO SG90 OU MG-996R                                    10/11/2020
;
; STATUS: FUNCIONA
;
; Controla a posicao de um servo motor SG90 ou MG996R atraves um potenciometro conectado a
; porta analogica RA0/AN0.
;
; Os valores dos tempos dos pulsos de controle do motor nao sao os ideais e foram
; recalculados, depois de varias tentativas, para fazer o motor girar 180 graus pois
; os valores teoricos so conseguem girar 90 graus.
; ---------------------------------------------------------------------------------------------


; ---------------------------------------------------------------------------------------------
; HARDWARE
; Fosc ........................... Cristal 4MHZ
; RA0/AN0 ........................ Porta analogica
; RC1 ............................ controle do servo motor 1
; ---------------------------------------------------------------------------------------------


; ---------------------------------------------------------------------------------------------
; CALCULEIRA
; --------------------------------------------------------------------------------------------- 
; 
; Os servos motores tabalham, em sua maioria, com pulsos numa frequencia se 50HZ no formato
; abaixo, onde a variacao de Time-On (Duty Cycle) e' responsavel pela rotacao do motor.
;  
;     +-------------+
;     |   Time-On   |              Time-Off             
;     |             |
;     |             +------------------------------------------------
;
;     |< ---------------------- Periodo = 20ms --------------------->|
;
; Teoricamente, um Time-On de 1,0ms posiciona o motor na extrema esquerda. 
; Time-On de 1,5ms posiciona o motor no centro e Time-On de 2,0ms posiciona
; o motor na extrema direita porem, isso nao vale na pratica. Certos motores,
; por deficiencia de construcao, podem funcionar com valores diferentes, como
; e' o caso dos motores usados neste projeto, o SG90 e o MG996R. 
; Sendo assim, os calculos abaixo aumentam a amplitude dos pulsos de controle,
; originalmente de 1000us a 2000us para 500us a 2500us mantendo a frequencia de 50HZ.
;
; Se o servo trabalha com frequencia de 50HZ, logo, seu periodo sera 
; T=1/50 --> T= 20000us  (mantivemos a frequencia teorica).
; 
; Fixemos os tempos de Time-On e Time-Off para a esquerda em microsegundos:
; TON = 500us  TOFF = T - TON = 20000 - 500
;              TOFF = 19500us
;
; Fixemos, agora, os tempos de Time-On e Time-Off para a direita em microsegundos:
; TON = 2500us TOFF = T - TON = 20000 - 2500
;              TOFF = 17500us
;
; Adotando a frequencia do clock (Fosc) do microcontrolador de 4MHZ, o ciclo de instrucao sera
; 1,0us e, se usarmos um prescaler 1:8 em TMR1, podemos calcular o valor de TMR1 como sendo:
; TMR = 65536 - (TON/8) para Time-On e  TMR = 65536 - (TOFF/8) para Time-Off. Assim temos:
;
;          TimeOn(us)  TMRTON             TimeOff(us)  TMRTOFF
;          ----------  ---------------    -----------  ----------------
; Esquerda  500        65536- 62=65474    19500        65536-2437=64099
; Direita  2500        65536-312=65224    17500        65536-2187=63349
;
;
; Para controlar o motor com um potenciometro conectado ao conversor analogico/digital,
; estabelecemos que:
; 1 - Com o potenciometro todo a esquerda (sentido anti-horario) o motor sera posicionado na 
;     extrema esquerda (visto por tras).
; 2 - Com o potenciometro todo a direita (sentido horario), o motor sera posicionado na 
;     extrema direita (visto por tras).
;
; Calculemos, agora, o passo a ser usado pelo ADC:
;                          
;        TMRON(esq) - TMRON(dir)    65474 - 65224     250
; Step = ----------------------- =  ------------- = ------ = 0,244
;                 1023                  1023         1023 
;
; -------------------------------------------------------------------------------------------------------


; -------------------------------------------------------------------------------------------------------
; Variaveis usadas para essa compilacao. 
; Os tempo sao calculados com base num Fosc de 4MHZ. 
; -------------------------------------------------------------------------------------------------------

 variable  fservo, tservo, prescaler, lton, ltoff, rton, srvLeftOn, srvLeftOff, srvRightOn, ADCStep

; Preencha os valores das variaveis abaixo se necessario
fservo      set  .50               ; HZ - frequencia do servo
prescaler   set  .8                ; prescaler do TMR1
lton        set  .500              ; microsegundos -> Time-On extrema esquerda
rton        set  .2580             ; microsegundos -> Time-On extrema direita

; Nao alterar daqui para baixo
tservo      set  .1000000/fservo   ; microsegundos -> periodo do servo
ltoff       set  tservo - lton     ; microsegundos -> Time-Off extrema esquerda
srvLeftOn   set  .65536 - (lton/prescaler)
srvLeftOff  set  .65536 - (ltoff/prescaler)
srvRightOn  set  .65536 - (rton/prescaler)
ADCStep     set  (.1000*(srvLeftOn - srvRightOn)) / .1023

; ---------------------------------------------------------------------------------------------
	LIST P=18F4520
	#include <P18F4520.INC>
    CONFIG WDT    = OFF
    CONFIG MCLRE  = ON
    CONFIG debug  = OFF
    CONFIG LVP    = OFF 
    CONFIG OSC    = HS               ; usa oscilador externo a cristal
;
; ---------------------------------------------------------------------------------------------
; MACROS
; --------------------------------------------------------------------------------------------- 
             #include  "familias.inc"
; ---------------------------------------------------------------------------------------------
; REGISTRADORES E SIMBOLOS NA RAM
; ---------------------------------------------------------------------------------------------
             UDATA  0x80

; PORTA DE CONTROLE DO MOTOR
#define       ENG1_IO     TRISC
#define       ENG1_PORT   PORTC
#define       ENG1_PIN    .1

ANRG1        res    .1                      ; MSB do valor lido no ADC
ANRG0        res    .1                      ; LSB do valor lido no ADC
TON1         res    .1                      ; Servo Time-On MSB
TON0         res    .1                      ; Time-On LSB
TOFF1        res    .1                      ; Time-Off MSB
TOFF0        res    .1                      ; Time-Off LSB

             #include  "gpr.inc"
             #include  "math.inc"

; ---------------------------------------------------------------------------------------------
; 
; ---------------------------------------------------------------------------------------------
rstvec     CODE    0x0000
           goto    inicio
inth       CODE    0x0008
           retfie 

; ---------------------------------------------------------------------------------------------
; 
; ---------------------------------------------------------------------------------------------
MAIN       CODE
           #include  "math_services.inc"

inicio     movlw    0x0F
           movwf    ADCON1                  ; todos os pino sao digitais  

; ---------------------------------------------------------------------------------------------
; Rotina principal
; ---------------------------------------------------------------------------------------------
           clrf     LATA
           clrf     LATC
           clrf     LATB
           clrf     TRISA
           clrf     TRISB
           clrf     TRISC

           call     settmr1                 ; configura TMR1
           call     confADC                 ; configura conversor analogico digital

anread     call     readADC                 ; le a porta analogica
           call     setTon                  ; calcula Time-On
           call     setToff                 ; calcula Time-Off
; ---------------------------------------------------------------------------------------------
engine1    movff    TON1, TMR1H             ; carrega MSB do TMR1 com Time ON
           movff    TON0, TMR1L             ; carrega LSB do TMR1 com Time-On
           bsf      ENG1_PORT, ENG1_PIN     ; levanta o sinal de controle do motor
           call     tstart                  ; conta tempo
           movff    TOFF1, TMR1H            ; carrega MSB do TMR1 com Time-Off
           movff    TOFF0, TMR1L            ; carrega MSB do TMR1 com Time-Off
           bcf      ENG1_PORT, ENG1_PIN     ; abaixa o sinal de controle do motor
           call     tstart                  ; conta tempo 
           goto     anread

; ---------------------------------------------------------------------------------------------
; Calculo do Time-Off do servo motor.
;
; TOFF = srvLeftOff + [ADC * (Step * 1000)]    onde
;
; srvLeftOff............. tempo de Time-Off a ser carregado em TMR1
; ADC ................... valor lido na porta analogica
; Step .................. passo calculado como Step=[(srvLeftOn - srvRightOn) / 1023] * 1000
;
; Para saber como os valores usados nas formulas foram deduzidos, ver comentarios em 
; 'CALCULEIRA' no inicio do programa.
;
; ---------------------------------------------------------------------------------------------
setToff    call     calcpart1
           movlw    high(srvLeftOff)
           movwf    REGA1
           movlw    low(srvLeftOff)
           movwf    REGA0
           call     add                     ; REGA = 65411 + (ADC * Step)
           movff    REGA1, TOFF1
           movff    REGA0, TOFF0
           return

; ---------------------------------------------------------------------------------------------
; Calculo do Time-On (Duty Cycle) do servo motor
;
; TON = srvLeftOn - [ADC * (Step * 1000)]    onde
;
; srvLeftOn ............. tempo de Time-On a ser carregado em TMR1
; ADC ................... valor lido na porta analogica
; Step .................. passo calculado como Step=[(srvLeftOn - srvRightOn) / 1023] * 1000
;
; Para saber como os valores usados nas formulas foram deduzidos, ver comentarios em 
; 'CALCULEIRA' no inicio do programa.
;
; ---------------------------------------------------------------------------------------------
setTon     call     calcpart1
           movlw    high(srvLeftOn)
           movwf    REGA1
           movlw    low(srvLeftOn)
           movwf    REGA0
           call     subtract                ; REGA = 65411 - (ADC * Step)
           movff    REGA1, TON1
           movff    REGA0, TON0
           return

; ---------------------------------------------------------------------------------------------
; Calcula primeira parcela [ADC*(Step*1000)] e retorna resultado em REGB
; ---------------------------------------------------------------------------------------------
calcpart1  call     clra
           call     clrb
           movff    ANRG1, REGA1
           movff    ANRG0, REGA0
           movlw    ADCStep 
           movwf    REGB0                   ; REGB = (Step * 1000)
           call     multiply                ; REGA = ADC * (Step * 1000)
           call     div_by10                ; divide REGA por 10
           call     div_by10                ; divide REGA por 100
           call     div_by10                ; divide REGA por 1000
           call     mv_a2b                  ; REGB = REGA
           call     clra
           return

; ---------------------------------------------------------------------------------------------
; REGA = REGA/10 --> Divide REGA por 10
; ---------------------------------------------------------------------------------------------
div_by10   call     clrb
           movlw    .10
           movwf    REGB0
           call     divide
           return
; ---------------------------------------------------------------------------------------------
; REGB = REGA --> Move REGA para REGB
; ---------------------------------------------------------------------------------------------
mv_a2b     movff    REGA3, REGB3
           movff    REGA2, REGB2
           movff    REGA1, REGB1
           movff    REGA0, REGB0
           return

; ---------------------------------------------------------------------------------------------
; Configura TRM1 como temporizador (T1CON<TMR1CS>=0) e Presc=8 (T1CON<5:4>=11)
; ---------------------------------------------------------------------------------------------
settmr1   movlw    0x30                     ; prescaler 1:8 T1CON<5:4>
          movwf    T1CON                    ; estabelece o prescaler
          bcf      PIR1,TMR1IF              ; limpa flag
          return

; ---------------------------------------------------------------------------------------------
; Aciona o timer
; ---------------------------------------------------------------------------------------------
tstart    bcf     PIR1, TMR1IF
          bsf     T1CON,TMR1ON  
tver      btfss   PIR1, TMR1IF
          goto    tver
          return

; ---------------------------------------------------------------------------------------------
; Configuracao da porta analogia
; ---------------------------------------------------------------------------------------------
confADC   bsf     TRISA,0                            ; RA0 entrada analogia
          clrf    ADCON0                             ; seleciona canal 0 (AN0)
          movlw   b'00001110'                        ; seleciona somente AN0
          movwf   ADCON1                             ; PCFG3:PCFG0 ---> DDDDDDDDDDDDA
          movlw   b'00010100'                        ; bits<5:3> --> 4 Tad     bits<3:0> --> Fosc/4               
          movwf   ADCON2
          bsf     ADCON2,ADFM                        ; right justify
          return
; ---------------------------------------------------------------------------------------------
; Leitura da porta analogia
; ---------------------------------------------------------------------------------------------
readADC   bsf     ADCON0,ADON                        ; conversor habilitado
          bsf     ADCON0,GO_DONE                     ; iniciar conversao
convwait  btfsc   ADCON0,GO_DONE                     ; espera a conversao terminar
          goto    convwait                           ; aguarda a conversao terminar
          movf    ADRESH,w                           ; obtem o MSB da conversao
          andlw   0x03                               ; deixa so bits 1 e 0 do MSB
          movwf   ANRG1                              ; armazena valor convertido
          movff   ADRESL,ANRG0                       ; armazena LSB do valor convertido
          return

          END 

